// $Id$
/**
* Copyright (C) 2016 EDIT
* European Distributed Institute of Taxonomy
* http://www.e-taxonomy.eu
*
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/
package eu.etaxonomy.taxeditor.molecular.handler;


import info.bioinfweb.jphyloio.JPhyloIOEventWriter;
import info.bioinfweb.jphyloio.ReadWriteParameterMap;
import info.bioinfweb.jphyloio.ReadWriteParameterNames;
import info.bioinfweb.jphyloio.dataadapters.implementations.ListBasedDocumentDataAdapter;
import info.bioinfweb.jphyloio.factory.JPhyloIOReaderWriterFactory;
import info.bioinfweb.jphyloio.objecttranslation.ObjectTranslatorFactory;

import java.io.File;
import java.io.IOException;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.TreeNode;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.ui.handlers.HandlerUtil;

import eu.etaxonomy.cdm.model.molecular.Sequence;
import eu.etaxonomy.taxeditor.editor.EditorUtil;
import eu.etaxonomy.taxeditor.model.MessagingUtils;
import eu.etaxonomy.taxeditor.molecular.TaxeditorMolecularPlugin;
import eu.etaxonomy.taxeditor.molecular.io.CDMPherogramAlignmentObjectTranslator;
import eu.etaxonomy.taxeditor.molecular.io.CDMSequenceMatrixAdapter;
import eu.etaxonomy.taxeditor.molecular.io.SingleReadAlignmentRDFXMLConstants;
import eu.etaxonomy.taxeditor.molecular.io.wizard.ExportSingleReadAlignmentWizard;
import eu.etaxonomy.taxeditor.molecular.l10n.Messages;
import eu.etaxonomy.taxeditor.util.ApplicationUtil;



/**
 * Allows to export a single read alignment to various alignment formats using
 * <a href="http://bioinfweb.info/JPhyloIO/"><i>JPhyloIO</i></a>.
 *
 * @author Ben Stöver
 * @date 24.04.2016
 */
public class ExportSequenceToFileHandler extends AbstractHandler {
    private static final JPhyloIOReaderWriterFactory factory = new JPhyloIOReaderWriterFactory();


    @Override
    public Object execute(ExecutionEvent event) throws ExecutionException {
        ISelection currentSelection = HandlerUtil.getCurrentSelection(event);
        TreeNode treeNodeOfSelection = EditorUtil.getTreeNodeOfSelection(currentSelection);
        if (treeNodeOfSelection != null && treeNodeOfSelection.getValue() instanceof Sequence) {
            Sequence sequence = (Sequence)treeNodeOfSelection.getValue();

            final ExportSingleReadAlignmentWizard wizard = new ExportSingleReadAlignmentWizard();
            final WizardDialog dialog = new WizardDialog(HandlerUtil.getActiveShell(event), wizard);
            if (dialog.open() == IStatus.OK) {
                // Prepare writer parameters:
                ReadWriteParameterMap parameters = new ReadWriteParameterMap();
                parameters.put(ReadWriteParameterNames.KEY_APPLICATION_NAME, ApplicationUtil.getTitle());
                //parameters.put(ReadWriteParameterNames.KEY_APPLICATION_VERSION, ApplicationUtil.getVersion());  // Setting the version unnecessary, since its already contained in the title.
                parameters.put(ReadWriteParameterNames.KEY_APPLICATION_URL, "http://cybertaxonomy.eu/taxeditor/");  //TODO Specify URL obtained from a central class? //$NON-NLS-1$
                parameters.put(ReadWriteParameterNames.KEY_SEQUENCE_EXTENSION_TOKEN, wizard.getModel().getElongationToken());

                // Create and register object translator for writing pherogram alignment shifts:
                ObjectTranslatorFactory translatorFactory = new ObjectTranslatorFactory();
                translatorFactory.addXSDTranslators(true);
                translatorFactory.addTranslator(new CDMPherogramAlignmentObjectTranslator(), true, SingleReadAlignmentRDFXMLConstants.DATA_TYPE_PHERORAGM_ALIGNMENT);
                parameters.put(ReadWriteParameterNames.KEY_OBJECT_TRANSLATOR_FACTORY, translatorFactory);

                // Create writer and document adapters:
                JPhyloIOEventWriter writer = factory.getWriter(wizard.getModel().getFormatInfo().getFormatID());
                ListBasedDocumentDataAdapter document = new ListBasedDocumentDataAdapter();
                document.getMatrices().add(new CDMSequenceMatrixAdapter(sequence, wizard.getModel().getConsensusSequenceLabel(),
                        wizard.getModel().isExportConsensusSequence(), wizard.getModel().isExportSingleReads()));

                // Write document:
                File file = new File(wizard.getModel().getFileName());
                if (!file.exists() || MessagingUtils.confirmDialog(Messages.exportSequenceToFileHandlerOverwriteTitle, String.format(
                        Messages.exportSequenceToFileHandlerOverwriteText, file.getAbsolutePath()))) {

                    try {
                        writer.writeDocument(document, file, parameters);
                    }
                    catch (IOException e) {
                        e.printStackTrace();
                        MessagingUtils.errorDialog(Messages.exportSequenceToFileHandlerIOErrorTitle, this,
                                String.format(Messages.exportSequenceToFileHandlerIOErrorMessage,
                                        file.getAbsolutePath()), TaxeditorMolecularPlugin.PLUGIN_ID, e, false);
                    }
                }

            }
        }
        return null;
    }
}
